-- Import necessary modules
require "import"
import "android.widget.*"
import "android.view.*"
import "android.app.AlertDialog"
import "android.app.Dialog"
import "android.content.Context"
import "android.content.ClipboardManager"
import "android.content.ClipData"
import "android.content.Intent" -- Needed for sharing
import "java.io.*"
import "java.lang.System"
import "java.lang.Thread"
import "java.lang.Runnable"
import "java.lang.StringBuilder"
import "java.lang.Runtime"
import "android.util.Log" -- Optional: for debugging background tasks

local TAG = "TerminalScript" -- Tag for Logcat debugging

-- --- Examples Data (Expanded and User-Friendly) ---
local exampleCommands = {
  { command = "echo 'Hello there!'", description = "Make the phone say 'Hello there!' in the output." },
  { command = "ls -la /sdcard/", description = "Show all files and folders (even hidden ones) in your main storage." },
  { command = "pwd", description = "Tell you which folder you are currently 'in'." },
  { command = "id", description = "Show your user ID number and groups you belong to." },
  { command = "sleep 5", description = "Make the script take a 5-second nap (shows how things can run without freezing)." },
  { command = "ping -c 4 google.com", description = "Check if you can reach Google (sends 4 test signals)." },
  { command = "logcat -d -v brief *:W", description = "Show recent warning or error messages from the phone's system log." },
  { command = "getprop ro.product.model", description = "Find out the exact model name of your phone." },
  { command = "getprop ro.build.version.release", description = "Check the Android version number (like 10, 11, 12)." },
  { command = "uptime", description = "Show how long your phone has been turned on since the last restart." },
  { command = "df -h", description = "Check how much storage space is free on your phone (easy to read format)." },
  { command = "mkdir /sdcard/MyNewFolder", description = "Create a new empty folder named 'MyNewFolder' in your main storage." },
  { command = "touch /sdcard/MyNewFile.txt", description = "Create a new empty file named 'MyNewFile.txt' in your main storage." },
  { command = "cp /sdcard/MyNewFile.txt /sdcard/MyNewFolder/", description = "Copy 'MyNewFile.txt' into the 'MyNewFolder' folder." },
  { command = "mv /sdcard/MyNewFile.txt /sdcard/AnotherName.txt", description = "Rename 'MyNewFile.txt' to 'AnotherName.txt'." },
  { command = "rm /sdcard/AnotherName.txt", description = "Delete the file 'AnotherName.txt'. Be careful!" },
  { command = "rmdir /sdcard/MyNewFolder", description = "Delete the empty folder 'MyNewFolder'. Only works if it's empty!" },
  { command = "ps", description = "Show a list of apps and processes currently running." },
  { command = "top -n 1 -m 5", description = "Show the top 5 running apps using the most phone power right now." },
  { command = "dumpsys battery", description = "Get detailed information about your phone's battery status." },
  { command = "svc wifi enable", description = "Turn WiFi ON using the system service command." },
  { command = "svc wifi disable", description = "Turn WiFi OFF using the system service command." },
  { command = "svc data enable", description = "Turn Mobile Data ON (might need special permission)." },
  { command = "svc data disable", description = "Turn Mobile Data OFF (might need special permission)." },
  { command = "settings get system screen_brightness", description = "Check the current screen brightness level (a number)." },
  -- { command = "settings put system screen_brightness 150", description = "Set screen brightness (needs special permission!)." }, -- Commented out: risky command example
  { command = "reboot", description = "Restart your phone (requires root!). Use with caution!" },
  { command = "whoami", description = "Show the username you are currently running commands as." },
  { command = "free -h", description = "Show how much memory (RAM) is being used and how much is free." },
  { command = "date", description = "Show the current date and time." },
}

-- --- Helper Functions ---

-- Function to trim leading/trailing whitespace from a string using Lua
local function trimString(s)
  if s == nil then return "" end
  return s:gsub("^%s+", ""):gsub("%s+$", "")
end

-- Function to show a simple alert dialog
local function showDialog(title, message)
  activity.runOnUiThread(Runnable{ -- Ensure dialogs are shown on UI thread
    run=function()
      AlertDialog.Builder(activity)
      .setTitle(title)
      .setMessage(message)
      .setPositiveButton("OK", nil)
      .show()
    end
  })
end

-- Function to copy text to clipboard
local function copyToClipboard(text)
  local clipboard = activity.getSystemService(Context.CLIPBOARD_SERVICE)
  if clipboard == nil then
     Toast.makeText(activity, "Clipboard service not available", Toast.LENGTH_SHORT).show()
     return
  end
  local clip = ClipData.newPlainText("Terminal Command", text)
  clipboard.setPrimaryClip(clip)
  Toast.makeText(activity, "Command copied to clipboard", Toast.LENGTH_SHORT).show()
end

-- Function to show the examples dialog
local function showExamplesDialog()
  local items = {}
  local commandsOnly = {}
  for i, ex in ipairs(exampleCommands) do
    local desc = ex.description or "No description"
    table.insert(items, ex.command .. "\n  (" .. desc .. ")")
    table.insert(commandsOnly, ex.command)
  end

  AlertDialog.Builder(activity)
  .setTitle("Example Commands (" .. #items .. ")")
  .setItems(items, function(dialog, which)
    local selectedCommand = commandsOnly[which + 1]
    if selectedCommand then
      copyToClipboard(selectedCommand)
      if editText then editText.setText(selectedCommand) end
    end
  end)
  .setNegativeButton("Cancel", nil)
  .show()
end

-- Function to share text using Android's share intent
local function shareText(text)
    if not text or text == "" then
        Toast.makeText(activity, "Nothing to share", Toast.LENGTH_SHORT).show()
        return
    end
    local intent = Intent(Intent.ACTION_SEND)
    intent.setType("text/plain")
    intent.putExtra(Intent.EXTRA_TEXT, text)
    activity.startActivity(Intent.createChooser(intent, "Share Output via..."))
end

-- --- Internal Command Execution Logic (used by executeCommand and checkRootStatus) ---
-- Returns: output, error, exitCode
local function runShellCommand(commandToRun)
    local outputStr = ""
    local errorStr = ""
    local exitCode = -1
    local process = nil

    local ok, result = pcall(function()
        local runtime = Runtime.getRuntime()
        Log.d(TAG, "Executing: " .. commandToRun) -- Log command being run
        process = runtime.exec({"/system/bin/sh", "-c", commandToRun})

        -- Non-blocking readers might be better, but simpler approach first:
        local outputStream = process.getInputStream()
        local outputReader = BufferedReader(InputStreamReader(outputStream))
        local outputBuilder = StringBuilder()
        local line = outputReader.readLine()
        while line ~= nil do
            outputBuilder.append(line).append("\n")
            line = outputReader.readLine()
        end
        outputReader.close()
        outputStr = outputBuilder.toString()

        local errorStream = process.getErrorStream()
        local errorReader = BufferedReader(InputStreamReader(errorStream))
        local errorBuilder = StringBuilder()
        line = errorReader.readLine()
        while line ~= nil do
            errorBuilder.append(line).append("\n")
            line = errorReader.readLine()
        end
        errorReader.close()
        errorStr = errorBuilder.toString()

        exitCode = process.waitFor()
        Log.d(TAG, "Command finished. Exit code: " .. exitCode) -- Log completion
    end)

    if not ok then
        errorStr = "Execution Error: " .. tostring(result)
        exitCode = -99 -- Indicate pcall failure
        if process then process.destroy() end
        Log.e(TAG, "pcall failed: " .. errorStr) -- Log the error
    end

    return trimString(outputStr), trimString(errorStr), exitCode
end


-- --- Asynchronous Command Execution ---
function executeCommand(command, asRoot)
  if not outputView then print("Error: outputView is nil!"); return end
  outputView.setText("Executing: " .. command .. (asRoot and " (as root)" or "") .. "\nPlease wait...")

  Thread(Runnable{
    run=function()
      local finalOutput, finalError, exitCode
      local isRootFailure = false
      local actualCommand = command

      if asRoot then
        actualCommand = "su -c '" .. command:gsub("'", "'\\''") .. "'"
      end

      -- Use the internal runner function
      finalOutput, finalError, exitCode = runShellCommand(actualCommand)

      -- Check for root failure *specifically* if we tried to run as root
      if asRoot and exitCode ~= 0 then
          -- More robust check: common errors OR specific exit codes like 127 (not found)
          if finalError:lower():match("permission denied") or
             finalError:lower():match("not allowed") or
             finalError:lower():match("su:") or
             finalError:match("exit code 127") or -- Often means 'su' not found
             exitCode == 1 or exitCode == 255 -- Common failure codes from su
          then
              isRootFailure = true
              Log.w(TAG, "Detected root failure. Exit: "..exitCode..", Error: "..finalError)
          end
      end

      -- Update UI on the Main Thread
      activity.runOnUiThread(Runnable{
         run=function()
            if not outputView then return end -- Check view again

            if isRootFailure then
                showDialog("Root Execution Failed", "Could not get root permission.\n\nReason: " .. (finalError ~= "" and finalError or "Unknown (Exit Code: " .. exitCode .. ")") .. "\n\nIs your device rooted and permission granted?")
                outputView.setText("Root execution failed (Exit Code: " .. exitCode .. "). See dialog.")
            elseif finalError ~= "" then
                outputView.setText("Exit Code: " .. exitCode .. "\nERROR:\n" .. finalError .. (finalOutput ~= "" and "\n\nOUTPUT:\n" .. finalOutput or ""))
            elseif finalOutput ~= "" then
                outputView.setText("Exit Code: " .. exitCode .. "\nOUTPUT:\n" .. finalOutput)
            else
                outputView.setText("Exit Code: " .. exitCode .. "\n(Command ran, but produced no output)")
            end
        end
      })
    end
  }).start()
end

-- --- Root Checker Function ---
function checkRootStatus()
    if not outputView then print("Error: outputView is nil!"); return end
    outputView.setText("Checking root status, please wait...")

    Thread(Runnable{
        run=function()
            local isLikelyRooted = false
            local evidence = StringBuilder()
            local suPath = ""

            -- 1. Try executing 'id' via 'su'
            evidence.append("--- Attempting 'su -c id' ---\n")
            local idOutput, idError, idExitCode = runShellCommand("su -c id")
            evidence.append("Exit Code: ").append(idExitCode).append("\n")
            if idExitCode == 0 and idOutput:match("uid=0(root)") then
                isLikelyRooted = true
                evidence.append("Result: Success! Found uid=0(root).\n")
                Log.i(TAG, "Root Check: 'su -c id' succeeded with root uid.")
            else
                evidence.append("Output: ").append(idOutput ~= "" and idOutput or "(none)").append("\n")
                evidence.append("Error: ").append(idError ~= "" and idError or "(none)").append("\n")
                evidence.append("Result: Failed or uid was not 0.\n")
                Log.w(TAG, "Root Check: 'su -c id' failed or no root uid. Exit: "..idExitCode..", Error: "..idError)
            end

            -- 2. If first check failed, try finding 'su' binary path
            if not isLikelyRooted then
                evidence.append("\n--- Attempting 'which su' ---\n")
                local whichOutput, whichError, whichExitCode = runShellCommand("which su")
                evidence.append("Exit Code: ").append(whichExitCode).append("\n")
                if whichExitCode == 0 and whichOutput ~= "" then
                    suPath = whichOutput:gsub("\n", "") -- Get path, remove newline
                    evidence.append("Output: ").append(suPath).append("\n")
                    evidence.append("Result: 'su' binary found at '").append(suPath).append("'. Device might be rooted, but access might require confirmation.\n")
                    Log.i(TAG, "Root Check: 'which su' found binary at: " .. suPath)
                    -- We don't set isLikelyRooted = true here, just note its presence.
                else
                    evidence.append("Output: ").append(whichOutput ~= "" and whichOutput or "(none)").append("\n")
                    evidence.append("Error: ").append(whichError ~= "" and whichError or "(none)").append("\n")
                    evidence.append("Result: 'su' binary not found in PATH.\n")
                    Log.w(TAG, "Root Check: 'which su' failed. Exit: "..whichExitCode..", Error: "..whichError)

                    -- 3. As a fallback, check common locations (less reliable)
                    evidence.append("\n--- Checking common locations for 'su' ---\n")
                    local lsOutput, lsError, lsExitCode = runShellCommand("ls /system/xbin/su /system/bin/su /su/bin/su")
                    evidence.append("Exit Code: ").append(lsExitCode).append("\n")
                    evidence.append("Output: ").append(lsOutput ~= "" and lsOutput or "(none)").append("\n")
                    evidence.append("Error: ").append(lsError ~= "" and lsError or "(none)").append("\n")
                    if lsExitCode == 0 and lsOutput:match("/su") then
                       evidence.append("Result: Found 'su' in a common location. Potential root.\n")
                       Log.i(TAG, "Root Check: Found 'su' via direct ls.")
                    else
                       evidence.append("Result: Did not find 'su' in common locations.\n")
                       Log.w(TAG, "Root Check: Did not find 'su' via direct ls.")
                    end
                end
            end

            -- Final Decision and Reporting
            local finalTitle, finalMessage
            if isLikelyRooted then
                finalTitle = "Root Access Detected!"
                finalMessage = "Congrats! It looks like your device is rooted and access was granted.\n\nYou should be able to run commands using the 'Execute as Root' button."
            elseif suPath ~= "" then
                finalTitle = "Potential Root Detected"
                finalMessage = "The 'su' binary was found at [" .. suPath .. "], but running a command as root didn't automatically succeed.\n\nYour device might be rooted, but you may need to grant permission via a root management app (like Magisk or SuperSU) when prompted."
            else
                finalTitle = "Root Access Not Detected"
                finalMessage = "Sorry, it doesn't look like your device is rooted, or the 'su' binary couldn't be found or executed.\n\nThe 'Execute as Root' button likely won't work."
            end

            -- Show result dialog on UI thread
            showDialog(finalTitle, finalMessage)
            -- Also update the text view with the detailed evidence
            activity.runOnUiThread(Runnable{ run=function()
                if outputView then outputView.setText("-- Root Check Details --\n" .. evidence.toString()) end
            end})
        end
    }).start()
end


-- --- Create Layout ---
layout = {
    LinearLayout,
    orientation = "vertical",
    layout_width = "fill",
    layout_height = "fill",
    padding = "5dp", -- Reduced padding slightly
    { -- Input Area
        EditText,
        id = "editText",
        layout_width = "fill",
        layout_height = "wrap_content",
        hint = "Enter command or pick from Examples",
        inputType = "textMultiLine|textNoSuggestions",
        textSize="14sp",
        padding="5dp",
    },
    { -- Main Button Row
        LinearLayout,
        orientation = "horizontal",
        layout_width = "fill",
        layout_height = "wrap_content",
        { -- Examples Button
            Button,
            id = "examplesButton",
            text = "Examples",
            layout_width = "0dp",
            layout_height = "wrap_content",
            layout_weight = 1,
            textSize="12sp",
            onClick = showExamplesDialog
        },
        { -- Execute Button
            Button,
            id = "executeButton",
            text = "Execute",
            layout_width = "0dp",
            layout_height = "wrap_content",
            layout_weight = 1,
            textSize="12sp",
            onClick = function()
                if not editText then print("Error: editText is nil!"); return end
                local command = trimString(editText.getText().toString())
                if command == "" then
                    if outputView then outputView.setText("Type a command first!") end
                else
                    executeCommand(command, false)
                end
            end
        },
        { -- Execute as Root Button
            Button,
            id = "executeRootButton",
            text = "Exec. Root", -- Abbreviated text
            layout_width = "0dp",
            layout_height = "wrap_content",
            layout_weight = 1,
            textSize="12sp",
            onClick = function()
                 if not editText then print("Error: editText is nil!"); return end
                 local command = trimString(editText.getText().toString())
                 if command == "" then
                     if outputView then outputView.setText("Type a command first!") end
                 else
                    executeCommand(command, true)
                 end
            end
        },
         { -- Check Root Button << NEW
            Button,
            id = "checkRootButton",
            text = "Check Root",
            layout_width = "0dp",
            layout_height = "wrap_content",
            layout_weight = 1,
            textSize="12sp",
            onClick = checkRootStatus -- Call the root check function
        }
    },
    { -- Second Button Row (New Features) << NEW
        LinearLayout,
        orientation = "horizontal",
        layout_width = "fill",
        layout_height = "wrap_content",
        paddingTop="4dp",
         { -- Clear Output Button
            Button,
            id = "clearButton",
            text = "Clear Output",
            layout_width = "0dp",
            layout_height = "wrap_content",
            layout_weight = 1,
            textSize="12sp",
            onClick = function()
                if outputView then
                    outputView.setText("") -- Simple clear
                    Toast.makeText(activity, "Output cleared", Toast.LENGTH_SHORT).show()
                end
            end
        },
        { -- Share Output Button
            Button,
            id = "shareButton",
            text = "Share Output",
            layout_width = "0dp",
            layout_height = "wrap_content",
            layout_weight = 1,
            textSize="12sp",
            onClick = function()
                if outputView then
                    shareText(outputView.getText().toString()) -- Call share helper
                else
                     Toast.makeText(activity, "Output view not found", Toast.LENGTH_SHORT).show()
                end
            end
        },
    },
    { -- Output Area (Scrollable)
      ScrollView,
      id="scroller", -- Give the scrollview an ID
      layout_width="fill",
      layout_height="0dp",
      layout_weight=1,
      fillViewport=true, -- Helps nested weights/heights
      {
        TextView,
        id = "outputView",
        layout_width = "fill",
        layout_height = "wrap_content", -- Let it grow inside scrollview
        text = "Welcome! Enter a command or select one from 'Examples'.",
        textSize="12sp",
        padding="5dp",
        textIsSelectable=true,
        -- gravity="bottom" -- Can sometimes cause issues with scrolling, keep default top
      }
    }
}

-- Set the layout
activity.setContentView(loadlayout(layout))

-- --- Post-Layout Setup ---
if not editText then print("ERROR: Could not find view with id 'editText'") end
if not outputView then print("ERROR: Could not find view with id 'outputView'") end
if not scroller then print("ERROR: Could not find view with id 'scroller'") end -- Check scroller too

-- Optional: Request focus
if editText then editText.requestFocus() end

-- Optional: Auto-scroll output to bottom after execution (might need refinement)
-- This is a bit tricky with async updates. A simpler approach might be a manual scroll button.
-- Or try focusing the bottom after setting text:
-- local function scrollToBottom()
--    if scroller and outputView then
--       scroller.post(Runnable{ run=function() scroller.fullScroll(View.FOCUS_DOWN) end })
--    end
-- end
-- Modify runOnUiThread in executeCommand/checkRootStatus to call scrollToBottom() after outputView.setText()

print("Script Loaded. UI Enhanced. Ready.")